"""
Library for combined x-ray and neutrons simulations.
====================================================
The neutron simulations is capable of handling non-magnetic, magnetic
non-spin flip as well as neutron spin-flip reflectivity. The model works
with scattering lengths densities directly.

The model is equivalent to spec_nx in calculation but uses different
parameterization that is more suitable for soft-matter applications.

Classes
-------

"""

from dataclasses import dataclass, field, fields
from typing import List

import numpy as np

# import all special footprint functions
from .lib import footprint as footprint_module
from .lib import neutron_refl as MatrixNeutron
from .lib import paratt as Paratt
from .lib import refl_base as refl
from .lib import resolution as resolution_module
from .lib.base import AltStrEnum
from .lib.footprint import *
from .lib.instrument import *
from .lib.physical_constants import muB_to_SL, r_e
from .lib.resolution import *
from .lib.testing import ModelTestCase
from .spec_nx import (AA_to_eV, Coords, FootType, Instrument, Polarization, Probe, ResType, footprintcorr, q_limit,
                      resolution_init, resolutioncorr)

# Preamble to define the parameters needed for the models outlined below:

ModelID = "SoftNX"

__xlabel__ = "q [Å$^{-1}$]"
__ylabel__ = "Instnsity [a.u.]"


@dataclass
class Layer(refl.ReflBase):
    """
    Representing a layer in the sample structur.

    ``d``
       The thickness of the layer in AA (Angstroms = 1e-10m)
    ``sigma``
       The root mean square roughness of the top interface of the layer in
       Angstroms.
    ``sld_x``
       The x-ray scattering length density in 1e-6 1/AA^2
    ``sld_n``
       The neutron scattering length density in 1e-6 1/AA^2
    ``sld_m``
       The neutron magnetic scattering length density in 1e-6 1/AA^2
    ``magn_ang``
       The angle of the magnetic moment in degress. 0 degrees correspond to
       a moment collinear with the neutron spin. Only used if probe is neutron spin-flip.
    """

    sigma: float = 0.0
    d: float = 0.0
    sld_n: complex = 1e-20j
    sld_x: complex = 0j
    sld_m: float = 0.0
    magn_ang: float = 0.0

    Units = {
        "sigma": "AA",
        "d": "AA",
        "sld_n": "1e-6 1/AA^2",
        "sld_x": "1e-6 1/AA^2",
        "sld_m": "1e-6 1/AA^2",
        "magn_ang": "deg.",
    }

    Groups = [("General", ["d", "sigma"]), ("Neutron", ["sld_n", "sld_m", "magn_ang"]), ("X-Ray", ["sld_x"])]

    DEFAULT_FIT_PARAMS = ["d", "sld_n", "sigma"]

@dataclass
class LayerParameters:
    sigma: List[float]
    d: List[float]
    sld_n: List[complex]
    sld_x: List[complex]
    sld_m: List[float]
    magn_ang: List[float]


@dataclass
class Stack(refl.StackBase):
    """
    A collection of Layer objects that can be repeated.

    ``Layers``
       A ``list`` consiting of ``Layers`` in the stack the first item is
       the layer closest to the bottom
    ``Repetitions``
       The number of repsetions of the stack
    """

    Layers: List[Layer] = field(default_factory=list)
    Repetitions: int = 1


@dataclass
class Sample(refl.SampleBase):
    """
    Describe global sample by listing ambient, substrate and layer parameters.

    ``Stacks``
       A ``list`` consiting of ``Stacks`` in the stacks the first item is
       the layer closest to the bottom
    ``Ambient``
       A ``Layer`` describing the Ambient (enviroment above the sample).
       Only the scattering lengths and density of the layer is used.
    ``Substrate``
       A ``Layer`` describing the substrate (enviroment below the sample).
       Only the scattering lengths, density and roughness of the layer is
       used.
    ``crop_sld``
       For samples with many layers this limits the number of layers that
       are shown on SLD graphs. Negative values just remove these layers
       and show a spike in SLD, instead, while postivie values exchange
       them by one layer of same thickness.
    """

    Stacks: List[Stack] = field(default_factory=list)
    Ambient: Layer = field(default_factory=Layer)
    Substrate: Layer = field(default_factory=Layer)
    crop_sld: int = 0
    _layer_parameter_class = LayerParameters


# A buffer to save previous calculations for spin-flip calculations
class Buffer:
    Ruu = 0
    Rdd = 0
    Rdu = 0
    Rud = 0
    parameters = None
    TwoThetaQz = None


def Specular(TwoThetaQz, sample: Sample, instrument: Instrument):
    """Simulate the specular signal from sample when probed with instrument

    # BEGIN Parameters
    TwoThetaQz data.x
    # END Parameters
    """
    global __xlabel__
    __xlabel__ = "q [Å$^{-1}$]"
    if instrument.coords == Coords.tth:
        __xlabel__ = "2θ [°]"

    # preamble to get it working with my class interface
    restype = instrument.restype
    Q, TwoThetaQz, weight = resolution_init(TwoThetaQz, instrument)
    # often an issue with resolution etc. so just replace Q values < q_limit
    Q = maximum(Q, q_limit)

    ptype = instrument.probe
    pol = instrument.pol

    parameters: LayerParameters = sample.resolveLayerParameters()

    if ptype == Probe.xray:
        # fb = array(parameters['f'], dtype = complex64)
        e = AA_to_eV / instrument.wavelength
        sld = refl.cast_to_array(parameters.sld_x, e) * 1e-6
    else:
        sld = array(parameters.sld_n, dtype=complex128) * 1e-6

    d = array(parameters.d, dtype=float64)
    sld_m = array(parameters.sld_m, dtype=float64) * 1e-6
    # Transform to radians
    magn_ang = array(parameters.magn_ang, dtype=float64) * pi / 180.0

    sigma = array(parameters.sigma, dtype=float64)

    l2pi = instrument.wavelength**2 / 2 / 3.141592
    # Ordinary Paratt X-rays
    if ptype == Probe.xray:
        R = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - l2pi * sld, d, sigma)
    # Ordinary Paratt Neutrons
    elif ptype == Probe.neutron:
        R = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - l2pi * sld, d, sigma)
    # Ordinary Paratt but with magnetization
    elif ptype == Probe.npol:
        # Polarization uu or ++
        if pol == Polarization.up_up:
            R = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - l2pi * (sld + sld_m), d, sigma)
        # Polarization dd or --
        elif pol == Polarization.down_down:
            R = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - l2pi * (sld - sld_m), d, sigma)
        elif pol == Polarization.asymmetry:
            Rp = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - l2pi * (sld - sld_m), d, sigma)
            Rm = Paratt.ReflQ(Q, instrument.wavelength, 1.0 - l2pi * (sld + sld_m), d, sigma)
            R = (Rp - Rm) / (Rp + Rm)
        else:
            raise ValueError("The value of the polarization is WRONG." " It should be uu(0) or dd(1)")
    # Spin flip
    elif ptype == Probe.npolsf:
        # Check if we have calcluated the same sample previous:
        if Buffer.TwoThetaQz is not None:
            Q_ok = Buffer.TwoThetaQz.shape == Q.shape
            if Q_ok:
                Q_ok = not (Buffer.TwoThetaQz != Q).any()
        else:
            Q_ok = False
        if Buffer.parameters != parameters or not Q_ok:
            if sld_m[-1] != 0.0 or sld[-1] != 0:
                sld_m -= sld_m[-1]
                sld -= sld[-1]
            sld_p = l2pi * (sld + sld_m)
            sld_m = l2pi * (sld - sld_m)
            Vp = (2 * pi / instrument.wavelength) ** 2 * (
                sld_p * (2.0 + sld_p)
            )  # (1-np**2) - better numerical accuracy
            Vm = (2 * pi / instrument.wavelength) ** 2 * (sld_m * (2.0 + sld_m))  # (1-nm**2)
            (Ruu, Rdd, Rud, Rdu) = MatrixNeutron.Refl(Q, Vp, Vm, d, magn_ang, sigma)
            Buffer.Ruu = Ruu
            Buffer.Rdd = Rdd
            Buffer.Rud = Rud
            Buffer.parameters = parameters
            Buffer.TwoThetaQz = Q.copy()
        else:
            pass
        if pol == Polarization.up_up:
            R = Buffer.Ruu
        elif pol == Polarization.down_down:
            R = Buffer.Rdd
        elif pol in [Polarization.up_down, Polarization.down_up]:
            R = Buffer.Rud
        # Calculating the asymmetry ass
        elif pol == Polarization.asymmetry:
            R = (Buffer.Ruu - Buffer.Rdd) / (Buffer.Ruu + Buffer.Rdd + 2 * Buffer.Rud)
        else:
            raise ValueError("The value of the polarization is WRONG." " It should be uu(0), dd(1) or ud(2)")
    # TODO: Check the following to cases
    # tof
    elif ptype == Probe.ntof:
        ai = instrument.incangle
        # if ai is an array, make sure it gets repeated for every resolution point
        if type(ai) is ndarray and restype in [ResType.full_conv_rel, ResType.full_conv_abs]:
            ai = (ai * ones(instrument.respoints)[:, newaxis]).flatten()
        else:
            ai = ai * ones(Q.shape)
        wl = 4 * pi * sin(ai * pi / 180) / Q
        l2pi = wl**2 / 2 / 3.141592
        R = Paratt.Refl_nvary2(ai, wl, 1.0 - l2pi * sld[:, newaxis], d, sigma, return_int=True)
    # tof spin polarized
    elif ptype == Probe.ntofpol:
        wl = 4 * pi * sin(instrument.incangle * pi / 180) / Q
        l2pi = wl**2 / 2 / 3.141592
        msld = sld_m[:, newaxis] * (4 * pi * sin(instrument.incangle * pi / 180) / Q) ** 2 / 2 / pi
        # polarization uu or ++
        if pol == Polarization.up_up:
            R = Paratt.Refl_nvary2(
                instrument.incangle * ones(Q.shape),
                (4 * pi * sin(instrument.incangle * pi / 180) / Q),
                1.0 - l2pi * (sld[:, newaxis] + msld),
                d,
                sigma,
                return_int=True,
            )
        # polarization dd or --
        elif pol == Polarization.down_down:
            R = Paratt.Refl_nvary2(
                instrument.incangle * ones(Q.shape),
                (4 * pi * sin(instrument.incangle * pi / 180) / Q),
                1.0 - l2pi * (sld[:, newaxis] - msld),
                d,
                sigma,
                return_int=True,
            )
        # Calculating the asymmetry
        elif pol == Polarization.asymmetry:
            Rd = Paratt.Refl_nvary2(
                instrument.incangle * ones(Q.shape),
                (4 * pi * sin(instrument.incangle * pi / 180) / Q),
                1.0 - l2pi * (sld[:, newaxis] - msld),
                d,
                sigma,
                return_int=True,
            )
            Ru = Paratt.Refl_nvary2(
                instrument.incangle * ones(Q.shape),
                (4 * pi * sin(instrument.incangle * pi / 180) / Q),
                1.0 - l2pi * (sld[:, newaxis] + msld),
                d,
                sigma,
                return_int=True,
            )
            R = (Ru - Rd) / (Ru + Rd)

        else:
            raise ValueError("The value of the polarization is WRONG." " It should be uu(0) or dd(1) or ass")
    else:
        raise ValueError("The choice of probe is WRONG")
    # FootprintCorrections

    foocor = footprintcorr(Q, instrument)
    # Resolution corrections
    R = resolutioncorr(R, TwoThetaQz, foocor, instrument, weight)

    return R * instrument.I0 + instrument.Ibkg


def EnergySpecular(Energy, TwoThetaQz, sample: Sample, instrument: Instrument):
    """Simulate the specular signal from sample when probed with instrument. Energy should be in eV.

    # BEGIN Parameters
    Energy data.x
    TwoThetaQz 3.0
    # END Parameters
    """
    # preamble to get it working with my class interface
    restype = instrument.restype
    # TODO: Fix so that resolution can be included.
    if restype != 0 and restype != ResType.none:
        raise ValueError("Only no resolution is allowed for energy scans.")
    if instrument.probe != Probe.xray:
        raise NotImplementedError("EnergySpecular only implemented for xray, for neutron ToF use Specular.")

    wl = AA_to_eV / Energy
    global __xlabel__
    __xlabel__ = "E [eV]"

    # TTH values given as x
    if instrument.coords == Coords.tth:
        theta = (TwoThetaQz + instrument.tthoff) / 2.0 * ones_like(Energy)
    # Q vector given....
    elif instrument.coords == Coords.q:
        theta = ((arcsin(TwoThetaQz * wl / 4 / pi) * 180.0 / pi) + instrument.tthoff / 2.0) * ones_like(Energy)

    parameters: LayerParameters = sample.resolveLayerParameters()
    sld = refl.cast_to_array(parameters.sld_x, Energy) * 1e-6

    d = array(parameters.d, dtype=float64)
    sigma = array(parameters.sigma, dtype=float64)

    l2pi = wl**2 / 2 / 3.141592
    # Ordinary Paratt X-rays
    R = Paratt.Refl_nvary2(theta, wl, 1.0 - l2pi * sld[:, newaxis], d, sigma)

    # TODO: Fix corrections
    # FootprintCorrections
    # foocor = footprintcorr(Q, instrument)
    # Resolution corrections
    # R = resolutioncorr(R, TwoThetaQz, foocor, instrument, weight)

    return R * instrument.I0 + instrument.Ibkg


def OffSpecular(TwoThetaQz, ThetaQx, sample, instrument):
    """Function that simulates the off-specular signal (not implemented)

    # BEGIN Parameters
    TwoThetaQz 1.0
    ThetaQx data.x
    # END Parameters
    """
    raise NotImplementedError("Not implemented use model interdiff insteads")


def SLD_calculations(z, item, sample: Sample, inst: Instrument):
    """Calculates the scatteringlength density as at the positions z
    if item is None or "all" the function returns a dictonary of values.
    Otherwise it returns the item as identified by its string.

    # BEGIN Parameters
    z data.x
    item 'Re'
    # END Parameters
    """
    parameters: LayerParameters = sample.resolveLayerParameters()
    # f = array(parameters['f'], dtype = complex64)
    e = AA_to_eV / inst.wavelength
    sld_x = refl.cast_to_array(parameters.sld_x, e)
    sld_n = array(parameters.sld_n, dtype=complex64)
    ptype = inst.probe
    magnetic = False
    mag_sld = 0
    sld_unit = "10^{-6}\\AA^{2}"
    if ptype == Probe.xray:
        sld = sld_x
    elif ptype == Probe.neutron:
        sld = sld_n
        sld_unit = "10^{-6}/\\AA^{2}"
    else:
        magnetic = True
        sld = sld_n
        sld_m = array(parameters.sld_m, dtype=float64)
        # Transform to radians
        magn_ang = array(parameters.magn_ang, dtype=float64) * pi / 180.0
        mag_sld = sld_m
        mag_sld_x = mag_sld * cos(magn_ang)
        mag_sld_y = mag_sld * sin(magn_ang)
        sld_unit = "10^{-6}/\\AA^{2}"

    d = array(parameters.d, dtype=float64)
    d = d[1:-1]
    # Include one extra element - the zero pos (substrate/film interface)
    int_pos = cumsum(r_[0, d])
    sigma = array(parameters.sigma, dtype=float64)[:-1] + 1e-7
    if z is None:
        z = arange(-sigma[0] * 5, int_pos.max() + sigma[-1] * 5, 0.5)
    if not magnetic:
        rho = sum((sld[:-1] - sld[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)), 1) + sld[-1]
        dic = {"Re": real(rho), "Im": imag(rho), "z": z, "SLD unit": sld_unit}
    else:
        sld_p = sld + mag_sld
        sld_m = sld - mag_sld
        rho_p = (
            sum((sld_p[:-1] - sld_p[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)), 1)
            + sld_p[-1]
        )
        rho_m = (
            sum((sld_m[:-1] - sld_m[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)), 1)
            + sld_m[-1]
        )
        rho_nucl = (rho_p + rho_m) / 2.0
        if inst.probe == Probe.npolsf and (magn_ang != 0.0).any():
            rho_mag_x = (
                sum(
                    (mag_sld_x[:-1] - mag_sld_x[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)),
                    1,
                )
                + mag_sld_x[-1]
            )
            rho_mag_y = (
                sum(
                    (mag_sld_y[:-1] - mag_sld_y[1:]) * (0.5 - 0.5 * erf((z[:, newaxis] - int_pos) / sqrt(2.0) / sigma)),
                    1,
                )
                + mag_sld_y[-1]
            )
            dic = {
                "Re non-mag": real(rho_nucl),
                "Im non-mag": imag(rho_nucl),
                "mag": real(rho_p - rho_m) / 2,
                "z": z,
                "mag_x": rho_mag_x,
                "mag_y": rho_mag_y,
                "SLD unit": sld_unit,
            }
        else:
            dic = {
                "Re non-mag": real(rho_nucl),
                "Im non-mag": imag(rho_nucl),
                "mag": real(rho_p - rho_m) / 2,
                "z": z,
                "SLD unit": sld_unit,
            }
    if item is None or item == "all":
        return dic
    else:
        try:
            return dic[item]
        except:
            raise ValueError("The chosen item, %s, does not exist" % item)


SimulationFunctions = {
    "Specular": Specular,
    "OffSpecular": OffSpecular,
    "SLD": SLD_calculations,
    "EnergySpecular": EnergySpecular,
}

Sample.setSimulationFunctions(SimulationFunctions)


class TestSoftNX(ModelTestCase):
    # TODO: currently this only checks for raise conditions in the code above, check of results should be added

    def test_spec_xray(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, sld_x=3e-5 + 1e-7j)])],
            Ambient=Layer(),
            Substrate=Layer(sld_x=5e-5 + 2e-7j),
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=1.54,
        )
        with self.subTest("x-ray tth"):
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray q"):
            instrument.coords = Coords.q
            Specular(self.qz, sample, instrument)

        # resolution corrections
        with self.subTest("x-ray q-res-fast"):
            instrument.restype = ResType.fast_conv
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-fast-rel"):
            instrument.restype = ResType.fast_conv_rel
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-fast-var"):
            instrument.restype = ResType.fast_conv_var
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-full"):
            instrument.restype = ResType.full_conv_abs
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-full-rel"):
            instrument.restype = ResType.full_conv_rel
            Specular(self.qz, sample, instrument)
        with self.subTest("x-ray q-res-wrong"):
            instrument.restype = 123
            with self.assertRaises(ValueError):
                Specular(self.qz, sample, instrument)
        instrument.restype = ResType.none

        # footprint corrections
        with self.subTest("x-ray q-footprint-square"):
            instrument.footype = FootType.square
            Specular(self.qz, sample, instrument)
        instrument.coords = Coords.tth
        with self.subTest("x-ray tth-footprint-square"):
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-gauss"):
            instrument.footype = FootType.gauss
            Specular(self.tth, sample, instrument)
        with self.subTest("x-ray tth-footprint-wrong"):
            instrument.footype = 123
            with self.assertRaises(ValueError):
                Specular(self.qz, sample, instrument)

    def test_spec_neutron(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, sld_n=3e-6, sld_m=1e-6)])],
            Ambient=Layer(sld_n=1e-7),
            Substrate=Layer(sld_n=4e-6),
        )
        instrument = Instrument(
            probe=Probe.neutron,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        with self.subTest("neutron tth"):
            Specular(self.tth, sample, instrument)
        with self.subTest("neutron q"):
            instrument.coords = Coords.q
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof q-footprint"):
            instrument.probe = Probe.ntof
            instrument.footype = FootType.square
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof q-footprint"):
            instrument.footype = FootType.none
            instrument.restype = ResType.full_conv_rel
            Specular(self.qz, sample, instrument)
            instrument.incangle = ones_like(self.qz) * 0.5
            Specular(self.qz, sample, instrument)
        instrument.incangle = 0.5
        instrument.restype = ResType.none
        with self.subTest("neutron-pol++ q"):
            instrument.probe = Probe.npol
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-pol-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-polsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpol++ q"):
            instrument.probe = Probe.ntofpol
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpol-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tofpolsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)

        with self.subTest("neutron tthoffset"):
            instrument.probe = Probe.neutron
            instrument.tthoff = 0.1
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-tof tthoffset"):
            instrument.probe = Probe.ntof
            instrument.tthoff = 0.1
            Specular(self.qz, sample, instrument)

        with self.subTest("neutron-sf++ q"):
            instrument.probe = Probe.npolsf
            instrument.pol = Polarization.up_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf-- q"):
            instrument.pol = Polarization.down_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf+- q"):
            instrument.pol = Polarization.up_down
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sf-+ q"):
            instrument.pol = Polarization.down_up
            Specular(self.qz, sample, instrument)
        with self.subTest("neutron-sfsa q"):
            instrument.pol = Polarization.asymmetry
            Specular(self.qz, sample, instrument)

    def test_energy(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, sld_x=3e-6)])],
            Ambient=Layer(sld_x=1e-7),
            Substrate=Layer(sld_x=4e-6),
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        energy = linspace(8000.0, 5000.0, 20)

        with self.subTest("energy xray"):
            EnergySpecular(energy, 1.5, sample, instrument)
        with self.subTest("energy xray q"):
            instrument.coords = Coords.q
            EnergySpecular(energy, 1.5, sample, instrument)

    def test_sld(self):
        sample = Sample(
            Stacks=[Stack(Layers=[Layer(d=150, sigma=2.0, sld_x=2e-5 + 1e-7j, sld_n=3e-6, sld_m=1e-6, magn_ang=24.0)])],
            Ambient=Layer(sld_n=1e-7),
            Substrate=Layer(sld_n=4e-6),
        )
        instrument = Instrument(
            probe=Probe.xray,
            coords=Coords.tth,
            res=0.001,
            restype=ResType.none,
            beamw=0.1,
            footype=FootType.none,
            tthoff=0.0,
            wavelength=4.5,
            incangle=0.5,
        )
        with self.subTest("sld xray"):
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron"):
            instrument.probe = Probe.neutron
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron pol"):
            instrument.probe = Probe.npolsf
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron pol2"):
            sample.Stacks[0].Layers[0].magn_ang = 0.0
            SLD_calculations(None, None, sample, instrument)
        with self.subTest("sld neutron crop"):
            instrument.probe = Probe.neutron
            sample.crop_sld = 15
            sample.Stacks[0].Repetitions = 100
            SLD_calculations(None, None, sample, instrument)
            sample.crop_sld = -15
            sample.Stacks[0].Repetitions = 100
            SLD_calculations(None, None, sample, instrument)


def standard_xray():
    """
    return the defied standard x-ray reflectivity to compare against other models
    """
    qz = linspace(0.01, 0.3, 15)
    scl = 0.1 * AA_to_eV * 1.54 / 2 / pi
    return Specular(
        qz,
        Sample(
            Ambient=Layer(),
            Substrate=Layer(d=150.0, sld_x=(1e-5 + 1e-8j) * scl),
            Stacks=[Stack(Layers=[Layer(d=150.0, sld_x=(2e-5 + 2e-8j) * scl)])],
        ),
        Instrument(probe=Probe.xray, coords=Coords.q, wavelength=1.54, footype=FootType.none, restype=ResType.none),
    )



import argparse
import logging
import multiprocessing
import os
import os.path
import sys

from logging import debug, getLogger
from threading import Thread

import platformdirs

from . import version
from .core import custom_logging
from .core.custom_logging import activate_excepthook, activate_logging, iprint, setup_system

config_path = os.path.abspath(platformdirs.user_data_dir("GenX3", "ArturGlavic"))
os.environ["NUMBA_CACHE_DIR"] = os.path.join(config_path, "numba_cache")


def start_interactive(args):
    """
    Start genx in interactive mode (with the gui)

    :param args: command line arguments evaluated with argparse.
    """
    debug("enter start_interactive")
    activate_excepthook()
    # Check if single instance is running
    from .gui.single_instance import SingleInstanceClient
    sc = SingleInstanceClient()
    if sc.is_server_running():
        if args.infile:
            sc.send_message([args.infile])
        else:
            sc.send_message([])
        sys.exit(0)
    # Fix blurry text on Windows 10
    import ctypes

    try:
        ctypes.windll.shcore.SetProcessDpiAwareness(True)
    except:
        pass

    from .gui import main_window

    if args.infile != "":
        debug("start GUI setup with file to load")
        filename = os.path.abspath(args.infile)
        # if getattr(sys, 'frozen', False):
        # make sure the path is set to the program executable in a frozen binary
        # this fixes unnecessery re-compiles of numba functions on Windows
        #    os.chdir(os.path.dirname(os.path.abspath(sys.executable)))
        #    debug(f'changed curdir to {os.path.abspath(os.path.curdir)}')
    else:
        filename = None
    debug("start GUI setup")
    if args.debug:
        debug("initialize wx Inspection App")
        from wx.lib.mixins.inspection import InspectionMixin

        class InspApp(main_window.GenxApp, InspectionMixin):
            def OnInit(self):
                # Initialize the inspection tool.
                self.Init()
                return main_window.GenxApp.OnInit(self)

        app = InspApp(filename=filename, dpi_overwrite=args.dpi_overwrite)
    else:
        app = main_window.GenxApp(filename=filename, dpi_overwrite=args.dpi_overwrite)
    debug("setup complete, start WX MainLoop")
    app.MainLoop()
    debug("leave start_interactive")


def create_simulated_data(args):
    """Function to create simulated data from the model and add it to data.y"""
    from scipy.stats import poisson

    from .core import config as io
    from .diffev import DiffEv
    from .model_control import ModelController

    io.config.load_default(os.path.join(os.path.dirname(os.path.abspath(__file__)), "profiles", "default.profile"))
    ctrl = ModelController(DiffEv())
    mod = ctrl.model

    iprint("Loading file: %s " % args.infile)
    ctrl.load_file(args.infile)
    # io.load_opt_config(opt, config)
    iprint("File loaded")

    iprint("Simualting..")
    mod.simulate()
    iprint("Storing simualted data")
    for data_set in mod.data:
        data_set.y_raw = poisson.rvs(data_set.y_sim)
        data_set.y_command = "y"
        data_set.run_y_command()
        data_set.error_command = "sqrt(where(y > 0, y, 1.0))"
        data_set.run_error_command()

    iprint("Saving the model to %s" % args.outfile)
    ctrl.save_file(args.outfile)


def extract_parameters(args):
    """Extracts the parameters to outfile"""

    from .core import config as io
    from .diffev import DiffEv
    from .model_control import ModelController

    # Open the genx file
    io.config.load_default(os.path.join(os.path.dirname(os.path.abspath(__file__)), "profiles", "default.profile"))
    ctrl = ModelController(DiffEv())
    mod = ctrl.model

    iprint("Loading file: %s " % args.infile)
    ctrl.load_file(args.infile)
    iprint("File loaded")

    names, values = mod.parameters.get_sim_pars()
    index_names = mod.parameters.get_names()
    errors = [
        mod.parameters[index_names.index(ni)].error if mod.parameters.get_fit_state_by_name(ni) == 1 else "-"
        for ni in names
    ]
    names = [ni if mod.parameters.get_fit_state_by_name(ni) != 1 else ni + "*" for ni in names]

    paramstr = "\t".join([name for name in names]) + "\n" + "\t".join(["%f" % val for val in values]) + "\n"
    if args.error:
        paramstr += "\t".join([err for err in errors]) + "\n"
    if args.outfile != "":
        iprint(f"Appending parameter values to file {args.outfile}")
        with open(args.outfile, "a") as fout:
            # Add header
            fout.write(paramstr)
    iprint("\n" + paramstr)


def modify_file(args):
    """Modify a GenX file given command line arguments"""
    from .core import config as io
    from .diffev import DiffEv
    from .model_control import ModelController

    # Open the genx file
    io.config.load_default(os.path.join(os.path.dirname(os.path.abspath(__file__)), "profiles", "default.profile"))
    ctrl = ModelController(DiffEv())
    mod = ctrl.model

    iprint("Loading file: %s " % args.infile)
    ctrl.load_file(args.infile)
    iprint("File loaded")

    if args.datafile:
        datafile = os.path.abspath(args.datafile)
        if 0 > args.data_set or args.data_set >= len(mod.data):
            iprint("The selected data set does not exist - select one between 0 and %d" % (len(mod.data) - 1))
            return
        iprint("Loading dataset %s into data set %d" % (datafile, args.data_set))
        mod.data[args.data_set].loadfile(datafile)

        if args.outfile:
            iprint("Saving the fit to %s" % args.outfile)
            ctrl.save_file(args.outfile)

    elif args.save_datafile:
        save_datafile = os.path.abspath(args.save_datafile)
        iprint("Simualting..")
        mod.simulate()
        if 0 > args.data_set or args.data_set >= len(mod.data):
            iprint("The selected data set does not exist - select one between 0 and %d" % (len(mod.data) - 1))
            return
        iprint("Exporting data set %d into ASCII file %s" % (args.data_set, save_datafile))
        mod.data[args.data_set].save_file(save_datafile)


def set_numba_single():
    debug("Setting numba JIT compilation to single CPU")
    os.environ["NUMBA_CACHE_DIR"] = os.path.join(config_path, "single_cpu_numba_cache")

    import numba

    old_jit = numba.jit

    def jit(*args, **opts):
        opts["parallel"] = False
        return old_jit(*args, **opts)

    numba.jit = jit
    numba.GENX_OVERWRITE_SINGLE = True
    try:
        numba.set_num_threads(1)
    except AttributeError:
        pass


class InputThread(Thread):
    def __init__(self):
        Thread.__init__(self, daemon=True)
        self.stop_fit = False

    def run(self):
        # concole input has to be queried differently for non-UNIX systems
        if sys.platform.startswith("win"):
            self.run_windows()
        else:
            self.run_unix()

    def run_windows(self):
        import msvcrt
        import time

        while not self.stop_fit:
            if msvcrt.kbhit():
                debug("Key event received")
                key = msvcrt.getwch()
                if "q" in key:
                    self.stop_fit = True
            time.sleep(0.1)

    def run_unix(self):
        import fcntl
        import selectors
        import time

        # set sys.stdin non-blocking
        orig_fl = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin, fcntl.F_SETFL, orig_fl | os.O_NONBLOCK)

        m_selector = selectors.DefaultSelector()
        m_selector.register(sys.stdin, selectors.EVENT_READ)
        while True:
            (k, evt) = m_selector.select()[0]
            res = k.fileobj.read()
            if "q" in res:
                self.stop_fit = True
                break
        m_selector.close()


def start_fitting(args, rank=0):
    """
    Function to start fitting from the command line.
    """
    import time

    from .core import config as io
    from .core.console import calc_errorbars, setup_console
    from .model_control import ModelController

    # Open the genx file
    io.config.load_default(os.path.join(os.path.dirname(os.path.abspath(__file__)), "profiles", "default.profile"))
    if not args.bumps:
        from .diffev import DiffEv

        ctrl = ModelController(DiffEv())
        mod = ctrl.model
        opt: DiffEv = ctrl.optimizer
        set_optimiser_pars = set_diffev_pars
    else:
        from .bumps_optimizer import BumpsOptimizer

        ctrl = ModelController(BumpsOptimizer())
        mod = ctrl.model
        opt: BumpsOptimizer = ctrl.optimizer
        set_optimiser_pars = set_bumps_pars

    if rank == 0:
        setup_console(ctrl, args.error, args.outfile, use_curses=(args.use_curses and not args.mpi))

    if (args.mpi or args.pr > 0) and not args.disable_numba:
        try:
            set_numba_single()
        except ImportError:
            pass

    if rank == 0:
        iprint("Loading model %s..." % args.infile)
    ctrl.load_file(args.infile)
    # has to be used in order to save everything....
    if args.esave:
        io.config.set("solver", "save all evals", True)
    # Simulate, this will also compile the model script
    if rank == 0:
        iprint("Simulating model...")
    mod.simulate()

    # Sets up the fitting ...
    set_optimiser_pars(opt, args)
    if rank == 0:
        iprint("Setting up the optimizer...")
        iprint(opt)

    if args.outfile and rank == 0:
        iprint("Saving the initial model to %s" % args.outfile)
        ctrl.save_file(args.outfile)

    # To start the fitting
    if rank == 0:
        iprint("Fitting starting...")
        t1 = time.time()
    # print opt.use_mpi, opt.use_parallel_processing
    opt.start_fit(mod)
    if rank == 0:
        inp = InputThread()
        inp.start()
    else:

        class Inp:
            stop_fit = False

        inp = Inp()
    while opt.is_running():
        try:
            time.sleep(0.1)
            if inp.stop_fit:
                opt.stop = True
        except KeyboardInterrupt:
            iprint("KeyboardInterrupt, trying to stop fit.")
            opt.stop = True

    if rank == 0:
        t2 = time.time()
        iprint("Fitting finished!")
        iprint("Time to fit: ", (t2 - t1) / 60.0, " min")

    if rank == 0:
        iprint("Updating the parameters")
        mod.parameters.set_value_pars(opt.best_vec)

    if args.outfile and rank == 0:
        if args.error:
            iprint("Calculating errorbars")
            calc_errorbars(mod, opt)
        iprint("Saving the fit to %s" % args.outfile)
        # opt.set_use_mpi(False)
        ctrl.save_file(args.outfile)

    if rank == 0:
        iprint("Fitting successfully completed")


def set_diffev_pars(optimiser, args):
    """
    Sets the optimiser parameters from args
    """
    if args.pr:
        optimiser.set_processes(args.pr)
        optimiser.set_use_parallel_processing(True)

    if args.mpi:
        optimiser.set_use_mpi(True)
    else:
        optimiser.set_use_mpi(False)

    if args.cs:
        optimiser.set_chunksize(args.cs)

    if args.mgen:
        optimiser.set_max_generations(args.mgen)
        optimiser.set_use_max_generations(True)

    if args.pops:
        optimiser.set_pop_size(args.pops)
        optimiser.set_use_pop_mult(False)

    if args.asi:
        optimiser.set_autosave_interval(args.asi)
        optimiser.set_use_autosave(True)

    if args.var >= 0:
        optimiser.opt.min_parameter_spread = args.var

    if args.km >= 0:
        optimiser.set_km(args.km)

    if args.kr >= 0:
        optimiser.set_kr(args.kr)


def set_bumps_pars(optimiser, args):
    if args.pr:
        optimiser.opt.use_parallel_processing = True
        optimiser.opt.parallel_processes = args.pr


def compile_numba(cache_dir=None):
    try:
        # perform a compilation of numba functions with console feedback
        import numba

        from .models.lib.numba_integration import configure_numba

        configure_numba()

        import inspect

        real_jit = numba.jit

        class UpdateJit:
            update_counter = 1

            def __call__(self, *args, **opts):
                if inspect.stack()[1][3] != "<lambda>":
                    print(f"compiling numba functions {self.update_counter}/22")
                    self.update_counter += 1
                return real_jit(*args, **opts)

        print("Starting to compile numba functions..")
        numba.jit = UpdateJit()
        print("paratt")
        from .models.lib import paratt_numba

        print("neutron")
        from .models.lib import neutron_numba

        print("instrument")
        from .models.lib import instrument_numba

        print("offspec")
        from .models.lib import offspec

        print("surface")
        from .models.lib import surface_scattering

        numba.jit = real_jit
    except Exception as e:
        print("An exception occured when trying to compile the numba functions:")
        print(e)
        return 1
    return 0


def main():
    multiprocessing.set_start_method("spawn")
    multiprocessing.freeze_support()
    if os.path.abspath(__file__).startswith("/snap"):
        # try fix multiprocessing in SNAP
        snap_sem = "/snap.genx.mp"
        multiprocessing.current_process()._config["semprefix"] = snap_sem
    else:
        snap_sem = None
    # Attempt to load mpi:
    try:
        from mpi4py import MPI
    except ImportError:
        rank = 0
        __mpi__ = False
    else:
        __mpi__ = bool(MPI.COMM_WORLD.Get_size() > 1)
        rank = MPI.COMM_WORLD.Get_rank()

    parser = argparse.ArgumentParser(
        prog="genx",
        description="GenX %s, fits data to a model." % version.__version__,
        epilog="For support, manuals and bug reporting see http://genx.sf.net",
    )
    run_group = parser.add_mutually_exclusive_group()
    run_group.add_argument("-r", "--run", action="store_true", help="run GenX fit (no gui)")
    if __mpi__:
        run_group.add_argument("--mpi", action="store_true", help="run GenX fit with mpi (no gui)")
    run_group.add_argument("-g", "--gen", action="store_true", help="generate data.y with poisson noise added")
    run_group.add_argument("--pars", action="store_true", help="extract the parameters from the infile")
    run_group.add_argument("--mod", action="store_true", help="modify the GenX file")
    opt_group = parser.add_argument_group("optimization arguments")
    opt_group.add_argument("--pr", type=int, default=0, help="Number of processes used in parallel fitting.")
    opt_group.add_argument("--cs", type=int, default=0, help="Chunk size used for parallel processing.")
    opt_group.add_argument("--mgen", type=int, default=0, help="Maximum number of generations that is used in a fit")
    opt_group.add_argument("--pops", type=int, default=0, help="Population size - number of individuals.")
    opt_group.add_argument("--asi", type=int, default=0, help="Auto save interval (generations).")
    opt_group.add_argument("--km", type=float, default=-1, help="Mutation constant (float 0 < km < 1)")
    opt_group.add_argument("--kr", type=float, default=-1, help="Cross over constant (float 0 < kr < 1)")
    opt_group.add_argument("-s", "--esave", action="store_true", help="Force save evals to gx file.")
    opt_group.add_argument("-e", "--error", action="store_true", help="Calculate/export error bars.")
    opt_group.add_argument(
        "--var", type=float, default=-1, help="Minimum relative parameter variation to stop the fit (%%)"
    )
    opt_group.add_argument(
        "--bumps", action="store_true", help="Use Bumps DREAM optimizer instead of GenX Differential Evolution"
    )
    data_group = parser.add_argument_group("data arguments")
    data_group.add_argument(
        "-d", dest="data_set", type=int, default=0, help="Active data set to act upon. Index starting at 0."
    )
    data_group.add_argument("--load", dest="datafile", help="Load file into active data set. Index starting at 0.")
    data_group.add_argument("--export", dest="save_datafile", help="Save active data set to file. Index starting at 0.")
    data_group = parser.add_argument_group("startup options")
    data_group.add_argument(
        "-l", "--logfile", dest="logfile", default=None, type=str, help="Output debug information to logfile."
    )
    data_group.add_argument(
        "--debug",
        dest="debug",
        default=False,
        action="store_true",
        help="Show additional debug information on console/logfile",
    )
    data_group.add_argument(
        "--dpi-scale",
        dest="dpi_overwrite",
        default=None,
        type=float,
        help="Overwrite the detection of screen dpi scaling factor (=72/dpi)",
    )
    data_group.add_argument(
        "--no-curses",
        dest="use_curses",
        default=True,
        action="store_false",
        help="Disable Curses interactive console interface for command line " "fitting on UNIX systems.",
    )
    data_group.add_argument(
        "--disable-nb",
        dest="disable_numba",
        default=False,
        action="store_true",
        help="Disable the use of numba JIT compiler",
    )
    data_group.add_argument(
        "--nb1",
        dest="numba_single",
        default=False,
        action="store_true",
        help="Compile numba JIT functions without parallel computing support (use one core only). "
        "Caching in this case is done in a different user directory.",
    )
    data_group.add_argument(
        "--compile-nb",
        dest="compile_nb",
        default=False,
        action="store_true",
        help="Perform a first-/recompilation of the numba modules and exit",
    )

    parser.add_argument(
        "infile", nargs="?", default="", help="The .gx or .hgx file to load or .ort file to use as basis for model"
    )
    parser.add_argument("outfile", nargs="?", default="", help="The .gx  or hgx file to save into")

    args = parser.parse_args()
    if not __mpi__:
        args.mpi = False

    if args.run or args.mpi or args.pars or args.mod:
        # make sure at least info-messages are shown (default is warning)
        custom_logging.CONSOLE_LEVEL = min(logging.INFO, custom_logging.CONSOLE_LEVEL)
    if rank > 0:
        custom_logging.CONSOLE_LEVEL = logging.WARNING
    setup_system()
    if snap_sem:
        logging.debug(f"Multiprocessing was setup with semprefix for SNAP to {snap_sem}")

    if args.logfile:
        if __mpi__:
            lfbase, lfend = args.logfile.rsplit(".", 1)
            activate_logging(f"{lfbase}_{rank:02}.{lfend}")
        else:
            activate_logging(args.logfile)
    debug("Arguments from parser: %s" % args)
    if not args.outfile:
        args.outfile = args.infile
    args.outfile = os.path.abspath(args.outfile)
    if args.infile:
        args.infile = os.path.abspath(args.infile)

    if args.disable_numba:
        debug("disable numba")
        # set numba flag
        from genx.models import lib as modellib

        modellib.USE_NUMBA = False
    elif args.numba_single:
        set_numba_single()
    if args.compile_nb:
        sys.exit(compile_numba())

    if args.run:
        start_fitting(args)
    elif args.mpi:
        start_fitting(args, rank)
    elif args.gen:
        create_simulated_data(args)
    elif args.pars:
        extract_parameters(args)
    elif args.mod:
        modify_file(args)
    elif not args.run and not args.mpi:
        # Check if the application has been frozen
        if getattr(sys, "frozen", False):
            # Redirect all the output to log files
            log_file_path = platformdirs.user_log_dir("GenX3", "ArturGlavic")
            # Create dir if not found
            if not os.path.exists(log_file_path):
                os.makedirs(log_file_path)
            # print log_file_path
            # log_file_path = genx_gui._path + 'app_data/'
            sys.stdout = open(log_file_path + "/genx_out.log", "w")
            sys.stderr = open(log_file_path + "/genx_err.log", "w")
        start_interactive(args)


if __name__ == "__main__":
    main()



"""
DWBA extension for offspecular from rough surfaces.

Rebuild of old C++ extension module.
"""

import numba

from numpy import arcsin, arctan, complex128, conj, cumsum, fabs, float64, inf, real, zeros
from scipy import integrate
from scipy.special import factorial

from .elfield import *


@numba.jit(
    numba.float64[:](
        numba.float64[:],
        numba.complex128[:, :, :],
        numba.complex128[:, :, :],
        numba.float64,
        numba.float64,
        numba.float64[:],
        numba.float64[:],
        numba.complex128[:],
        numba.float64[:],
        numba.complex128[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64,
    ),
    nopython=True,
    parallel=True,
    cache=True,
)
def dwba_interdiff_sum(qx, G, q, eta, h, sigma, sigmaid, sqn, z, table, q_min, fn, eta_z):
    Layers = G.shape[1]
    Points = qx.shape[0]

    s = 0.0j
    I = zeros(qx.shape, dtype=float64)

    for m in numba.prange(Points):
        Itemp = 0.0j
        for i in range(Layers):
            for j in range(Layers):
                for k in range(4):
                    for l in range(4):
                        nind = (k, i, m)  # DWBA component, Layer, q-point for incident
                        npind = (l, j, m)  # DWBA component, Layer, q-point for outgoing
                        for p in range(fn.shape[0]):
                            xnew = abs(qx[m] * eta / pow(p + 1.0, 1.0 / 2.0 / h))
                            lower = int((xnew - q_min[0]) / (q_min[1] - q_min[0]))
                            s += (
                                2.0
                                * pow(
                                    q[nind] * conj(q[npind]) * sigma[i] * sigma[j] * exp(-abs(z[i] - z[j]) / eta_z),
                                    p + 1.0,
                                )
                                / fn[p]
                                * eta
                                / pow(p + 1.0, 1.0 / 2.0 / h)
                                * (
                                    (table[lower + 1] - table[lower])
                                    / (q_min[lower + 1] - q_min[lower])
                                    * (xnew - q_min[lower])
                                    + table[lower]
                                )
                            )
                        Itemp += (
                            (sqn[i] - sqn[i + 1])
                            * conj(sqn[j] - sqn[j + 1])
                            * G[nind]
                            * exp(-0.5 * pow(sigmaid[i] * q[nind], 2))
                            * conj(G[npind] * exp(-0.5 * pow(sigmaid[j] * q[npind], 2)))
                            * exp(-0.5 * (pow(q[nind] * sigma[i], 2) + pow(conj(q[npind]) * sigma[j], 2)))
                            / q[nind]
                            / conj(q[npind])
                            * s
                        )
                        s = 0
        I[m] = Itemp.real
    return I


@numba.jit(
    numba.float64[:](
        numba.float64[:],
        numba.complex128[:, :, :],
        numba.complex128[:, :, :],
        numba.float64,
        numba.float64,
        numba.float64[:],
        numba.complex128[:],
        numba.float64[:],
        numba.complex128[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64,
    ),
    nopython=True,
    parallel=True,
    cache=True,
)
def dwba_sum(qx, G, q, eta, h, sigma, sqn, z, table, q_min, fn, eta_z):
    Layers = G.shape[1]
    Points = qx.shape[0]
    s = 0.0j
    I = zeros(qx.shape, dtype=float64)

    for m in numba.prange(Points):
        Itemp = 0.0j
        for i in range(Layers):
            for j in range(Layers):
                for k in range(4):
                    for l in range(4):
                        nind = (k, i, m)  # DWBA component, Layer, q-point for incident
                        npind = (l, j, m)  # DWBA component, Layer, q-point for outgoing
                        for p in range(fn.shape[0]):
                            xnew = fabs(qx[m] * eta / pow(p + 1.0, 1.0 / 2.0 / h))
                            lower = int((xnew - q_min[0]) / (q_min[1] - q_min[0]))
                            s += (
                                2.0
                                * pow(
                                    q[nind] * conj(q[npind]) * sigma[i] * sigma[j] * exp(-abs(z[i] - z[j]) / eta_z),
                                    p + 1.0,
                                )
                                / fn[p]
                                * eta
                                / pow(p + 1.0, 1.0 / 2.0 / h)
                                * (
                                    (table[lower + 1] - table[lower])
                                    / (q_min[lower + 1] - q_min[lower])
                                    * (xnew - q_min[lower])
                                    + table[lower]
                                )
                            )
                        Itemp += (
                            (sqn[i] - sqn[i + 1])
                            * conj(sqn[j] - sqn[j + 1])
                            * G[nind]
                            * conj(G[npind])
                            * exp(-0.5 * (pow(q[nind] * sigma[i], 2.0) + pow(conj(q[npind]) * sigma[j], 2.0)))
                            / q[nind]
                            / conj(q[npind])
                            * s
                        )
                        s = 0.0j
        I[m] = Itemp.real
    return I


@numba.jit(
    numba.float64[:](
        numba.float64[:],
        numba.float64[:],
        numba.float64,
        numba.float64,
        numba.float64[:],
        numba.complex128[:],
        numba.float64[:],
        numba.complex128[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64,
    ),
    nopython=True,
    parallel=True,
    cache=True,
)
def born_ext(qx, qz, eta, h, sigma, sqn, z, table, q_min, fn, eta_z):
    Layers = sigma.shape[0]
    Points = qx.shape[0]
    s = 0.0j
    I = zeros(qx.shape, dtype=float64)

    for m in numba.prange(Points):
        Itemp = 0.0j
        for i in range(Layers):
            for j in range(Layers):
                for p in range(fn.shape[0]):
                    xnew = qx[m] * eta / pow(p + 1.0, 1.0 / 2.0 / h)
                    lower = int((xnew - q_min[0]) / (q_min[1] - q_min[0]))
                    s += (
                        2.0
                        * pow(qz[m] * qz[m] * sigma[i] * sigma[j] * exp(-abs(z[i] - z[j]) / eta_z), p + 1.0)
                        / fn[p]
                        * eta
                        / pow(p + 1.0, 1.0 / 2.0 / h)
                        * (
                            (table[lower + 1] - table[lower])
                            / (q_min[lower + 1] - q_min[lower])
                            * (xnew - q_min[lower])
                            + table[lower]
                        )
                    )

                Itemp += (
                    (sqn[i] - sqn[i + 1])
                    * conj(sqn[j] - sqn[j + 1])
                    * exp(-0.5 * (pow(qz[m] * sigma[i], 2.0) + pow(qz[m] * sigma[j], 2.0)))
                    / qz[m]
                    / qz[m]
                    * s
                )
                s = 0.0j
        I[m] = Itemp.real
    return I


def vec_realsymint(F, omega, eta, h, eta_z, qz_n, qz_np, sigma_n, sigma_np, max_n):
    I = 2 * sum(
        [
            ((qz_n * conj(qz_np) * (sigma_n * sigma_np)) ** n)
            / factorial(n)
            * eta
            / n ** (1.0 / 2.0 / h)
            * F(omega * eta / (n ** (1.0 / 2.0 / h)))
            for n in arange(1.0, max_n + 1.0)
        ]
    )
    I = array(I)
    return I


def make_F(q, h):
    def f(x):
        return exp(-abs(x) ** (2 * h))

    q = q * (abs(q) > 0.003) + 0.003 * (abs(q) < 0.003)
    I = [integrate.quad(f, 0, inf, weight="cos", wvar=x)[0] for x in q]
    I = array(I)
    return I


def DWBA_Interdiff(qx, qz, lamda, n, z, sigma, sigmaid, eta, h, eta_z, d=[0], taylor_n=1):
    # sigmaid-interdiffused part
    # sigma-roughness
    k = 2 * pi / lamda
    qx = array(qx, dtype=float64)
    # Calculating electrical fields
    omega = arctan(qx / qz)  # Not the real omega given by Fewster
    omegap = arcsin(sqrt(qx**2 + qz**2) / 2 / k)
    kxi = k * cos(omega + omegap)
    kxf = k * cos(omegap - omega)

    # Calculate the electric fields
    (T, R, kz) = AmpElfield_q(k, r_[kxi, kxf], lamda, n[::-1], r_[0, d[1:][::-1]])

    lenq = len(qx)
    Ri = R[:, :lenq]
    Rf = R[:, lenq:]
    Ti = T[:, :lenq]
    Tf = T[:, lenq:]
    ki = kz[:, :lenq]
    kf = kz[:, lenq:]
    G = array([Ti * Tf, Ti * Rf, Ri * Tf, Ri * Rf], dtype=complex128)
    q = array([ki + kf, ki - kf, -ki + kf, -ki - kf], dtype=complex128)

    # Setting up for the Fourier integral as given by Pape et.al.
    maxn = taylor_n
    # New dqmin tested 20051201
    dqmin = min(abs(qx[:-1] - qx[1:])) * eta / (maxn ** (1.0 / 2.0 / h))
    if abs(dqmin) < 1e-12:
        dqmin = 1e-3
    # New q_min tested 20051201
    q_min = arange(
        min(abs(qx)) * eta / (maxn ** (1.0 / 2.0 / h)) - dqmin, max(abs(qx)) * eta + 2 * dqmin, dqmin, dtype=float64
    )
    table = array(make_F(q_min, h), dtype=complex128)

    fn = factorial(arange(1, maxn + 1))
    sqn = n**2
    sqn = array(sqn, dtype=complex128)
    sigma = array(sigma, dtype=float64)
    sigmaid = array(sigmaid, dtype=float64)
    s = dwba_interdiff_sum(qx, G, q, eta, h, sigma, sigmaid, sqn, z, table, q_min, fn, eta_z)

    return (s, omega + omegap, omegap - omega)


def DWBA(qx, qz, lamda, n, z, sigma, eta, h, eta_z, d=[0], taylor_n=1):
    k = 2 * pi / lamda
    qx = array(qx, dtype=float64)
    sqn = n**2
    sqn = array(sqn, dtype=complex)
    # Calculating electrical fields
    omega = arctan(qx / qz)  # Not the real omega given by Fewster
    omegap = arcsin(sqrt(qx**2 + qz**2) / 2 / k)
    kxi = k * cos(omega + omegap)
    kxf = k * cos(omegap - omega)
    (T, R, kz) = AmpElfield_q(k, r_[kxi, kxf], lamda, n[::-1], r_[0, d[1:][::-1]])

    lenq = len(qx)
    Ri = R[:, :lenq]
    Rf = R[:, lenq:]
    Ti = T[:, :lenq]
    Tf = T[:, lenq:]
    ki = kz[:, :lenq]
    kf = kz[:, lenq:]
    G = array([Ti * Tf, Ti * Rf, Ri * Tf, Ri * Rf], dtype=complex)
    q = array([ki + kf, ki - kf, -ki + kf, -ki - kf], dtype=complex)

    # Setting up for the Fourier integral as given by Pape et.al.
    maxn = taylor_n
    dqmin = (qx[1] - qx[0]) * eta / (maxn ** (1.0 / 2.0 / h))
    if abs(dqmin) < 1e-12:
        dqmin = 1e-3
    q_min = arange(qx[0] * eta / (maxn ** (1.0 / 2.0 / h)), qx[-1] * eta + 2 * dqmin, dqmin, dtype=float64)

    table = array(make_F(q_min, h), dtype=complex)
    fn = factorial(arange(1, maxn + 1))
    s = dwba_sum(qx, G, q, eta, h, sigma, sqn, z, table, q_min, fn, eta_z)
    return (s, omega + omegap, omegap - omega)


def Born(qx, qz, lamda, n, z, sigma, eta, h, eta_z, d=[0], taylor_n=1):
    k = 2 * pi / lamda
    qx = array(qx, dtype=float64)
    sqn = n**2
    sqn = array(sqn, dtype=complex)
    # print 'Setup Complete'
    # Calculating electrical fields
    omega = arctan(qx / qz)  # Not the real omega given by Fewster
    omegap = arcsin(sqrt(qx**2 + qz**2) / 2 / k)
    # Setting up for the Fourier integral as given by Pape et.al.
    maxn = taylor_n
    dqmin = (qx[1] - qx[0]) * eta / (maxn ** (1.0 / 2.0 / h))
    q_min = arange(qx[0] * eta / (maxn ** (1.0 / 2.0 / h)), qx[-1] * eta + 2 * dqmin, dqmin, dtype=float64)

    table = make_F(q_min, h)
    table = array(table, dtype=complex128)

    fn = factorial(arange(1, maxn + 1))
    s = born_ext(qx, qz, eta, h, sigma, sqn, z, table, q_min, fn, eta_z)

    return (s, omega + omegap, omegap - omega)


if __name__ == "__main__":
    n = array([1] + [1 - 7.57e-6 + 1.73e-7j, 1 - 2.24e-5 + 2.89e-6j] * 10 + [1 - 7.57e-6 + 1.73e-7j])
    # n=array([1,1-2.25e-5+2.9e-6j,1-7.5e-6+1.75e-7j])
    # n=array([1,1-2.25e-5,1-7.5e-6])
    d = array([0.0] + [80.0, 20.0] * 10)
    # d=array([0.0,1000.0])
    z = -cumsum(d)
    sigma = array([5.0] + [5.0, 5.0] * 10)
    # sigma=array([3.0,3.0])

    qx = arange(0.1e-4, 0.005, 0.00002)
    qz = 0.2 * ones(qx.shape)

    # qx=arange(1e-5,3e-5,1e-5)
    eta = 200.0
    eta_z = 1.0
    h = 1.0

    (I1, ain, aout) = Born(qx, qz, 1.540562, n, z, sigma, eta, h, eta_z)
    (I2, ain, aout) = DWBA(qx, qz, 1.540562, n, z, sigma, eta, h, eta_z, d)
    (I3, ain, aout) = DWBA_Interdiff(qx, qz, 1.540562, n, z, sigma, 0.3 * sigma, eta, h, eta_z, d)

    from matplotlib import pyplot

    pyplot.plot(qx, real(I1), qx, real(I2), qx, real(I3))
    pyplot.show()



"""
Test of special list class in colors module.
"""

import unittest

from genx.core.colors import CyclicList


class TestCyclicList(unittest.TestCase):

    def test_cycling(self):
        t = CyclicList([1, 2, 3])
        for i in range(-20, 21):
            self.assertEqual(t[i], [1, 2, 3][i % 3])

    def test_eq_hash(self):
        t = CyclicList([1, 2, 3])

        self.assertEqual(t, t)
        self.assertEqual(t, [1, 2, 3])
        _ = {t: 1}



"""
Test all reflectivity models generically and discover all models implementing test cases.
"""

import os
import unittest

from numpy import testing

from genx import api
from genx.models.lib.testing import ModelTestCases
from glob import glob
from importlib import import_module

MODEL_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "genx", "models")

try:
    import numba
except ModuleNotFoundError:
    numba = None

class TestReflModels(unittest.TestCase):
    def test_empty_models(self):
        ref_refl = None
        models_to_test = ["spec_nx", "spec_adaptive", "spec_inhom", "soft_nx", "mag_refl"]
        if numba is not None:
            models_to_test += ["interdiff"]
        for i, name in enumerate(models_to_test):
            with self.subTest(f"Model {name} import", i=i):
                model, optimizer, refl = api.Reflectivity.create_new(name)
                refl.ReadModel()
                model.compile_script()
                model.simulate()
            if hasattr(model.script_module.model, 'standard_xray'):
                with self.subTest(f"Model {name} std x-ray"):
                    res = model.script_module.model.standard_xray()
                    if ref_refl is None:
                        ref_refl = res
                    else:
                        testing.assert_array_almost_equal(ref_refl, res)


def load_tests(loader, standard_tests, pattern):
    """
    Build a test suite for all model test cases. First import all models.* modules to define these tests.
    """
    for mi in glob(os.path.join(MODEL_DIR, "*.py")):
        module_name = os.path.basename(mi)[:-3]
        if mi in ['__init__', 'symmetries', 'utils']:
            continue
        try:
            import_module(f'genx.models.{module_name}', )
        except ModuleNotFoundError:
            raise unittest.SkipTest("Skipping module " + module_name + ", error on import")

    suite = unittest.TestSuite()
    for test_class in [TestReflModels]+ModelTestCases:
        tests = loader.loadTestsFromTestCase(test_class)
        suite.addTests(tests)
    return suite



"""
Plotting widget for sample SLDs.
"""

import colorsys
import os
import time

from dataclasses import dataclass

import numpy as np
import wx

from scipy.interpolate import interp1d

from genx.core.custom_logging import iprint
from genx.data import DataList
from genx.gui.plotpanel import BasePlotConfig, PlotPanel
from genx.model import Model

from .reflectivity_misc import ReflectivityModule


@dataclass
class SamplePlotConfig(BasePlotConfig):
    section = "sample plot"
    data_derived_color: bool = False
    show_single_model: bool = False
    legend_outside: bool = False

    use_mass_density: bool = False
    show_imag: bool = False


class SamplePlotPanel(PlotPanel):
    """
    Widget for plotting the scattering length density of a sample.
    """

    opt: SamplePlotConfig

    def __init__(self, parent, plugin, id=-1, color=None, dpi=None, style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        PlotPanel.__init__(self, parent, id, color, dpi, SamplePlotConfig, style, **kwargs)
        self.plugin = plugin

        self.update = self.Plot
        self.create_axes()
        self.plot_dicts = []

    def create_axes(self):
        self.ax = self.figure.add_subplot(111)
        box = self.ax.get_position()
        if self.opt.legend_outside:
            self.ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        else:
            self.ax.set_position([box.x0, box.y0, box.width * 0.95, box.height])

    def Plot(self):
        """
        Plotting the sample.
        """
        if not self.ax:
            # self.ax = self.figure.add_subplot(111)
            self.create_axes()

        data: DataList = self.plugin.GetModel().get_data()
        if self.opt.data_derived_color:
            colors = []
            for di in data:
                h, s, v = colorsys.rgb_to_hsv(*di.sim_color)
                ci = [
                    colorsys.hsv_to_rgb(h, 1.0, 1.0),
                    colorsys.hsv_to_rgb(h, 1.0, 0.5),
                    colorsys.hsv_to_rgb(h, 0.4, 1.0),
                    colorsys.hsv_to_rgb((h + 0.1) % 1.0, 1.0, 1.0),
                    colorsys.hsv_to_rgb((h + 0.1) % 1.0, 1.0, 0.5),
                    colorsys.hsv_to_rgb((h + 0.1) % 1.0, 0.4, 1.0),
                    colorsys.hsv_to_rgb((h - 0.1) % 1.0, 1.0, 1.0),
                    colorsys.hsv_to_rgb((h - 0.1) % 1.0, 1.0, 0.5),
                    colorsys.hsv_to_rgb((h - 0.1) % 1.0, 0.4, 1.0),
                ]
                colors.append(ci)
        else:
            colors = [["b", "r", "g", "c", "m", "y", "k"]] * len(data)

        model: ReflectivityModule = self.plugin.GetModel().script_module
        self.plot_dicts = []
        while len(self.ax.lines) > 0:
            self.ax.lines[0].remove()
        while len(self.ax.collections) > 0:
            self.ax.collections[0].remove()
        i = 0
        sld_units = []

        if self.plugin.sim_returns_sld and model._sim:
            # New style sim function with one sld for each simulation
            self.plot_dicts = list(model.SLD)
            for sim in range(len(self.plot_dicts)):
                if data[sim].show:
                    if self.opt.use_mass_density:
                        sld_dict = self.plot_dicts[sim].get('Mass Density', {})
                    else:
                        sld_dict = self.plot_dicts[sim]
                    for key in sld_dict:
                        try:
                            if key in ["z", "SLD unit", "Mass Density"] or (sld_dict[key] == 0).all():
                                # skip lines that are all zero to keep legend cleaner
                                continue
                        except KeyError:
                            pass
                        is_imag = key[:2] == "Im" or key[:4] == "imag"
                        if (is_imag and self.opt.show_imag) or not is_imag:
                            if self.opt.show_single_model:
                                label = key
                            else:
                                label = data[sim].name + "\n" + key
                            self.ax.plot(
                                sld_dict["z"],
                                sld_dict[key],
                                color=colors[sim][i % len(colors[sim])],
                                label=label,
                            )

                            if "SLD unit" in sld_dict:
                                if not sld_dict["SLD unit"] in sld_units:
                                    sld_units.append(sld_dict["SLD unit"])
                            i += 1
                if self.opt.show_single_model:
                    break
        else:
            # Old style plotting just one sld
            if self.plugin.GetModel().compiled:
                try:
                    sample = model.sample
                except AttributeError:
                    iprint("Warning: Could not locate the sample in the model")
                    return
                plot_dict = sample.SimSLD(None, None, model.inst)
                self.plot_dicts = [plot_dict]
                for key in self.plot_dicts[0]:
                    if key in ["z", "SLD unit", "Mass Density"] or (self.plot_dicts[0][key] == 0).all():
                        # skip lines that are all zero to keep legend cleaner
                        continue
                    is_imag = key[:2] == "Im" or key[:4] == "imag"
                    if (is_imag and self.opt.show_imag) or not is_imag:
                        label = key
                        self.ax.plot(
                            self.plot_dicts[0]["z"],
                            self.plot_dicts[0][key],
                            colors[-1][i % len(colors[-1])],
                            label=label,
                        )

                        if "SLD unit" in self.plot_dicts[0]:
                            if not self.plot_dicts[0]["SLD unit"] in sld_units:
                                sld_units.append(self.plot_dicts[0]["SLD unit"])
                        i += 1

        if i > 0:
            if self.opt.legend_outside:
                self.ax.legend(loc="center left", framealpha=0.5, bbox_to_anchor=(1, 0.5), fontsize="small", ncol=1)
            else:
                self.ax.legend(loc="upper right", framealpha=0.5, fontsize="small", ncol=1)
            sld_unit = ", ".join(sld_units)
            if self.opt.use_mass_density:
                self.ax.yaxis.label.set_text(r"$\mathrm{\mathsf{density\,[%s]}}$"%sld_unit)
            else:
                self.ax.yaxis.label.set_text(r"$\mathrm{\mathsf{SLD\,[%s]}}$" % sld_unit)
            self.ax.xaxis.label.set_text(r"$\mathrm{\mathsf{ z\,[\AA]}}$")
            wx.CallAfter(self.flush_plot)
            self.AutoScale()

    def SavePlotData(self, filename):
        """Save all the SLD profiles to file with filename."""
        # Check so that there are a simulation to save
        try:
            self.plot_dicts
        except:
            self.plugin.ShowWarningDialog("No SLD data to save." " Simulate the model first and then save.")
            return
        base, ext = os.path.splitext(filename)
        if ext == "":
            ext = ".dat"
        data = self.plugin.GetModel().get_data()
        for sim in range(len(self.plot_dicts)):
            new_filename = base + "%03d" % sim + ext
            save_array = np.array([self.plot_dicts[sim]["z"]])
            header = " z [Å]" + " " * 12
            for key in self.plot_dicts[sim]:
                if key == "Mass Density":
                    continue
                if key not in ["z", "SLD unit"]:
                    save_array = np.r_[save_array, [self.plot_dicts[sim][key]]]
                    header += f" {key:19s}"
            for key in self.plot_dicts[sim].get("Mass Density", {}):
                if key not in ["z", "SLD unit"]:
                    save_array = np.r_[save_array, [self.plot_dicts[sim]["Mass Density"][key]]]
                    header += f" MD_{key:16s}"

            with open(new_filename, "w", encoding="utf-8") as f:
                f.write("# File exported from GenX's Reflectivity plugin\n")
                f.write("# File created: %s\n" % time.ctime())
                f.write("# Simulated SLD for data set: %s\n" % data[sim].name)
                if "SLD unit" in self.plot_dicts[sim]:
                    sld_unit = self.plot_dicts[sim]["SLD unit"]
                    sld_unit = sld_unit.replace("10^{-6}\\AA^{-2}", "10⁻⁶ Å⁻²")
                    sld_unit = sld_unit.replace("r_{e}/\\AA^{3}", "rₑ/Å⁻³")
                    f.write("# SLD unit: %s\n" % sld_unit)
                if "SLD unit" in self.plot_dicts[sim].get("Mass Density", {}):
                    sld_unit = self.plot_dicts[sim]["Mass Density"]["SLD unit"]
                    f.write("# Mass Density (MD) unit: %s\n" % sld_unit)
                f.write("# Coumns: \n")
                f.write("#" + header + "\n")
                np.savetxt(f, save_array.T, fmt="%.13e")

    def generate_sld_distribution(self, reference_interface=0, number_sample=1000):
        """
        Create SLD confidence interval from a sample of random parameters chosen from
        the parameter errors. A reference surface defines the zero point for the comparison.
        """
        model_obj: Model = self.plugin.GetModel()
        parameters = model_obj.parameters
        model_module: ReflectivityModule = self.plugin.GetModel().script_module

        param_funcs, best_params, par_min, par_max = model_obj.get_fit_pars(use_bounds=False)
        param_edown, param_eup = np.array(parameters.get_error_pars()).T
        NP = len(param_funcs)

        initial_SLDs = model_module.SLD
        # setup empty lists for each element of the model SLD
        data = []
        zmin = 1e6
        zmax = -1e6
        for SLDi in initial_SLDs:
            data.append(dict([(key, []) for key in SLDi if key not in ["z", "SLD unit", "Mass Density"]]))
            zmin = min(SLDi["z"].min(), zmin)
            zmax = max(SLDi["z"].max(), zmax)
        # create a general z-range for all simulations
        zrng = zmax - zmin
        z = np.arange(round(zmin - 0.1 * zrng), round(zmax + 0.1 * zrng), 1.0)

        if reference_interface != 0:
            d = model_module.sample.resolveLayerParameters()["d"]
            h = np.cumsum(d)
            z -= h[reference_interface]
            for SLDi in initial_SLDs:
                SLDi["z"] -= h[reference_interface]

        for i in range(number_sample):
            srnd = np.random.randn(NP)
            dp = np.where(srnd > 0.0, param_eup, param_edown) * abs(srnd)
            pi = best_params + dp
            for j, fj in enumerate(param_funcs):
                fj(pi[j])
            model_module.SLD = []
            model_module.Sim(model_obj.data)

            for i, SLDi in enumerate(model_module.SLD):
                if reference_interface != 0:
                    d = model_module.sample.resolveLayerParameters()["d"]
                    h = np.cumsum(d)
                    SLDi["z"] -= h[reference_interface]

                for key, value in SLDi.items():
                    if key in ["z", "SLD unit", "Mass Density"]:
                        continue
                    zfun = interp1d(
                        SLDi["z"], value, fill_value=(value[0], value[-1]), bounds_error=False, kind="linear"
                    )
                    data[i][key].append(zfun(z))

        # calculate the SLD curve range that falls within the 1-sigma interval of 68.2% probability
        output = []
        for i, SLDi in enumerate(data):
            output.append({"z": z, "SLD unit": initial_SLDs[i]["SLD unit"]})
            for key, value in SLDi.items():
                zfun = interp1d(
                    initial_SLDs[i]["z"],
                    initial_SLDs[i][key],
                    fill_value=(initial_SLDs[i][key][0], initial_SLDs[i][key][-1]),
                    bounds_error=False,
                    kind="linear",
                )

                output[-1][key] = (
                    np.percentile(np.array(value), 2.5, axis=0),  # lower bound 2-sigma
                    np.percentile(np.array(value), 15.9, axis=0),  # lower bound 1-sigma
                    zfun(z),  # best parameter line
                    np.percentile(np.array(value), 84.1, axis=0),  # upper bound 1-sigma
                    np.percentile(np.array(value), 97.5, axis=0),
                )  # upper bound 2-sigma
        return output

    def PlotConfidence(self, reference_interface=0, number_sample=1000, plot_data=None):
        """
        Plotting the sample.
        """
        if not self.ax:
            # self.ax = self.figure.add_subplot(111)
            self.create_axes()

        data: DataList = self.plugin.GetModel().get_data()
        if self.opt.data_derived_color:
            colors = []
            for di in data:
                h, s, v = colorsys.rgb_to_hsv(*di.sim_color)
                ci = [
                    colorsys.hsv_to_rgb(h, 1.0, 1.0),
                    colorsys.hsv_to_rgb(h, 1.0, 0.5),
                    colorsys.hsv_to_rgb(h, 0.4, 1.0),
                    colorsys.hsv_to_rgb((h + 0.1) % 1.0, 1.0, 1.0),
                    colorsys.hsv_to_rgb((h + 0.1) % 1.0, 1.0, 0.5),
                    colorsys.hsv_to_rgb((h + 0.1) % 1.0, 0.4, 1.0),
                    colorsys.hsv_to_rgb((h - 0.1) % 1.0, 1.0, 1.0),
                    colorsys.hsv_to_rgb((h - 0.1) % 1.0, 1.0, 0.5),
                    colorsys.hsv_to_rgb((h - 0.1) % 1.0, 0.4, 1.0),
                ]
                colors.append(ci)
        else:
            colors = [["b", "r", "g", "c", "m", "y", "k"]] * len(data)

        while len(self.ax.lines) > 0:
            self.ax.lines[0].remove()
        while len(self.ax.collections) > 0:
            self.ax.collections[0].remove()

        if plot_data is None:
            plot_data = self.generate_sld_distribution(reference_interface, number_sample)
        i = 0
        sld_units = []

        # New style sim function with one sld for each simulation
        for sim in range(len(plot_data)):
            if data[sim].show:
                for key in plot_data[sim]:
                    try:
                        if key in ["z", "SLD unit"] or (plot_data[0][key][2] == 0).all():
                            # skip lines that are all zero to keep legend cleaner
                            continue
                    except KeyError:
                        pass
                    is_imag = key[:2] == "Im" or key[:4] == "imag"
                    if (is_imag and self.opt.show_imag) or not is_imag:
                        if self.opt.show_single_model:
                            label = key
                        else:
                            label = data[sim].name + "\n" + key
                        # 2-sigma range
                        self.ax.fill_between(
                            plot_data[sim]["z"],
                            plot_data[sim][key][0],
                            plot_data[sim][key][4],
                            color=colors[sim][i % len(colors[sim])],
                            alpha=0.25,
                        )
                        # 1-sigma range
                        self.ax.fill_between(
                            plot_data[sim]["z"],
                            plot_data[sim][key][1],
                            plot_data[sim][key][3],
                            color=colors[sim][i % len(colors[sim])],
                            alpha=0.5,
                        )
                        # fit result line
                        self.ax.plot(
                            plot_data[sim]["z"],
                            plot_data[sim][key][2],
                            color=colors[sim][i % len(colors[sim])],
                            label=label,
                        )

                        if "SLD unit" in plot_data[sim]:
                            if not plot_data[sim]["SLD unit"] in sld_units:
                                sld_units.append(plot_data[sim]["SLD unit"])
                        i += 1
            if self.opt.show_single_model:
                break

        if i > 0:
            if self.opt.legend_outside:
                self.ax.legend(loc="center left", framealpha=0.5, bbox_to_anchor=(1, 0.5), fontsize="small", ncol=1)
            else:
                self.ax.legend(loc="upper right", framealpha=0.5, fontsize="small", ncol=1)
            sld_unit = ", ".join(sld_units)
            self.ax.yaxis.label.set_text(r"$\mathrm{\mathsf{SLD\,[%s]}}$" % sld_unit)
            self.ax.xaxis.label.set_text(r"$\mathrm{\mathsf{ z\,[\AA]}}$")
            wx.CallAfter(self.flush_plot)
            self.AutoScale()

    def SaveConfidenceData(self, filename, reference_interface=0, number_sample=1000, do_plot=False):
        """Save all the SLD confidence intervals to file with filename."""
        plot_data = self.generate_sld_distribution(reference_interface, number_sample)
        base, ext = os.path.splitext(filename)
        if ext == "":
            ext = ".dat"
        data = self.plugin.GetModel().get_data()
        sub_names = ["-l95", "-l63", "-bst", "-h63", "-h95"]
        for sim in range(len(plot_data)):
            new_filename = base + "%03d" % sim + ext
            save_array = np.array([plot_data[sim]["z"]])
            header = " z [Å]" + " " * 12
            for key in plot_data[sim]:
                if key != "z" and key != "SLD unit":
                    for i, sn in enumerate(sub_names):
                        save_array = np.r_[save_array, [plot_data[sim][key][i]]]
                        header += f" {key+sn:19s}"
            with open(new_filename, "w", encoding="utf-8") as f:
                f.write("# File exported from GenX's Reflectivity plugin\n")
                f.write("# File created: %s\n" % time.ctime())
                f.write("# Simulated SLD with confidence intervals for data set: %s\n" % data[sim].name)
                f.write("# Reference interface/Number of Samples: %i/%i\n" % (reference_interface, number_sample))
                f.write(
                    "# Data describes confidence intervals given as 2-sigma (95%%), "
                    "1-sigma (68%%) and best estimate from fit."
                )
                if "SLD unit" in plot_data[sim]:
                    sld_unit = plot_data[sim]["SLD unit"]
                    sld_unit = sld_unit.replace("10^{-6}\\AA^{-2}", "10⁻⁶ Å⁻²")
                    sld_unit = sld_unit.replace("r_{e}/\\AA^{3}", "rₑ/Å⁻³")
                    f.write("# SLD unit: %s\n" % sld_unit)
                f.write("# Coumns: \n")
                f.write("#" + header + "\n")
                np.savetxt(f, save_array.T, fmt="%.13e")
        if do_plot:
            self.PlotConfidence(plot_data=plot_data)

    def OnContextMenu(self, event):
        """
        Callback to show the popmenu for the plot which allows various
        settings to be made.
        """
        menu = self.generate_context_menu()
        menu.AppendSeparator()

        legendID = wx.NewId()
        menu.AppendCheckItem(legendID, "Legend Outside")
        menu.Check(legendID, self.opt.legend_outside)

        def OnLegend(event):
            self.opt.legend_outside = not self.opt.legend_outside
            self.figure.clear()
            self.create_axes()
            self.Plot()

        self.Bind(wx.EVT_MENU, OnLegend, id=legendID)

        coloringID = wx.NewId()
        menu.AppendCheckItem(coloringID, "Color from Data")
        menu.Check(coloringID, self.opt.data_derived_color)

        def OnColoring(event):
            self.opt.data_derived_color = not self.opt.data_derived_color
            self.Plot()

        self.Bind(wx.EVT_MENU, OnColoring, id=coloringID)

        singlemodelID = wx.NewId()
        menu.AppendCheckItem(singlemodelID, "Single Sample Model")
        menu.Check(singlemodelID, self.opt.show_single_model)

        def OnSingleModel(event):
            self.opt.show_single_model = not self.opt.show_single_model
            self.Plot()

        self.Bind(wx.EVT_MENU, OnSingleModel, id=singlemodelID)

        # Time to show the menu
        self.PopupMenu(menu)
        self.WriteConfig()

        self.Unbind(wx.EVT_MENU)
        menu.Destroy()
